//! Control routines for the x86
//! 8259 Programmable Interrupt Controller
//!
//! This helps us configure interrupts that receive
//! keyboard input and timer pulses.

use crate::os::util::outb;

/// IO base address for master PIC
const PIC_A: u32 = 0x20;
/// Command address for master PIC
const PIC_A_COMMAND: u32 = PIC_A;
/// Data address for master PIC
const PIC_A_DATA: u32 = PIC_A + 1;

/// IO base address for slave PIC
const PIC_B: u32 = 0xA0;
/// Command address for slave PIC
const PIC_B_COMMAND: u32 = PIC_B;
/// Data address for slave PIC
const PIC_B_DATA: u32 = PIC_B + 1;

/// A driver for the PIC
///
/// Reference:
/// - https://wiki.osdev.org/8259_PIC
/// - https://os.phil-opp.com/hardware-interrupts
pub struct PICDriver {
	offset_pic_a: u8,
	offset_pic_b: u8,
}

impl PICDriver {
	/// Create a PIC driver with the given offsets
	pub const fn new(offset_pic_a: u8, offset_pic_b: u8) -> Self {
		Self {
			offset_pic_a,
			offset_pic_b,
		}
	}

	fn send_a_cmd(&self, cmd: u8) {
		unsafe { outb(PIC_A_COMMAND, cmd) }
	}

	fn send_a_data(&self, cmd: u8) {
		unsafe { outb(PIC_A_DATA, cmd) }
	}

	fn send_b_cmd(&self, cmd: u8) {
		unsafe { outb(PIC_B_COMMAND, cmd) }
	}

	fn send_b_data(&self, cmd: u8) {
		unsafe { outb(PIC_B_DATA, cmd) }
	}

	/// Send an EOI for the given IRQ.
	///
	/// This needs to be called at the end of each interrupt handler.
	/// If `both` is true, reset both PICs. This is only necessary
	/// when we handle interrupts from PIC_B.
	pub fn send_eoi(&self, both: bool) {
		if both {
			self.send_b_cmd(0x20);
		}
		self.send_a_cmd(0x20);
	}

	/// Initialize this PIC driver.
	/// This should be called as early as possible.
	pub fn init(&mut self) {
		const ICW1_ICW4: u8 = 0x01; /* Indicates that ICW4 will be present */
		const ICW1_INIT: u8 = 0x10; /* Initialization - required! */
		const ICW4_8086: u8 = 0x01; /* 8086/88 (MCS-80/85) mode */

		self.send_a_cmd(ICW1_INIT | ICW1_ICW4);
		self.send_b_cmd(ICW1_INIT | ICW1_ICW4);

		self.send_a_data(self.offset_pic_a); // ICW2: Master PIC vector offset
		self.send_b_data(self.offset_pic_b); // ICW2: Slave PIC vector offset

		self.send_a_data(4); // ICW3: tell Master PIC that there is a slave PIC at IRQ2 (0000 0100)
		self.send_b_data(2); // ICW3: tell Slave PIC its cascade identity (0000 0010)

		// ICW4: have the PICs use 8086 mode (and not 8080 mode)
		self.send_a_data(ICW4_8086);
		self.send_b_data(ICW4_8086);

		// Unmask both PICs
		self.send_a_data(0);
		self.send_b_data(0);
	}
}
